// ----------------------------------------------------------------------------
// Wave3.java
// 3-Wave Resonance Condition Solver
// $Id: Wave3.java,v 1.33 2009/07/03 11:15:30 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------

import java.util.*;

public final class Wave3
{
  // domain bound to be used for testing
  private final static int D=100;
  
  // print format 
  private final static int FORMATNONE = 0; // no output
  private final static int FORMATLINE = 1; // single line Mathematica format
  private final static int FORMATCSV = 2;  // comma separated values
  private final static int format = FORMATLINE;
  
  // which solutions are printed
  private final static int SOLUTIONSALL   =  0; // n1+n2=n3 or n1-n2=n3
  private final static int SOLUTIONSPLUS  = +1; // n1+n2=n3
  private final static int SOLUTIONSMINUS = -1; // n1-n2=n3
  
  /***************************************************************************
   * Program to solve resonance conditions for 3-wave interactions.
   * 
   * Call as "java -server Wave3 [ -s ] [ <D> ][ <OPT> ]"
   * 
   * Computes all solutions in domain { (m,n) | m>=0, n>=0, m^2+n^2 <= D^2 }
   * 
   * OPT =  0: solutions satisfying n1+n2=n3 or n1-n2=n3 are printed (default).
   * OPT =  1: only solutions satisfying n1+n2=n3 are printed.
   * OPT = -1: only solutions satisfying n1-n2=n3 are printed.
   *
   * Prints list of solutions to standard output stream; unless option "-s"
   * is given, also the number of solutions and the time for their 
   * computation is printed.
   * 
   * @param argv the argument vector described above.
   **************************************************************************/
  public static void main(String argv[])
  {
    // input processing
    boolean silent = false;
    int d = D;
    int sols = SOLUTIONSALL;
    int n = argv.length;
    int p = 0;
    try
    {
      if (p < n && argv[p].equals("-s"))
      {
        silent = true;
        p = p+1;
      }
      if (p < n)
      {
        d = Integer.parseInt(argv[p]);
        if (d < 0) usage();
        p = p+1;
      }
      if (p < n)
      {
        sols = Integer.parseInt(argv[p]);
        if (sols < -1 || sols > +1) usage();
      }
    }
    catch (NumberFormatException e)
    {
      usage();
      return;
    }
    
    // computation
    long start = System.currentTimeMillis();
    SolutionSet solutions = solve(d, sols);
    long end = System.currentTimeMillis();
    
    // output
    if (format == FORMATNONE)
      { }
    else if (format == FORMATLINE)
      solutions.print();
    else if (format == FORMATCSV)
      solutions.printCSV();
    if (!silent)
    {
      System.out.println("Solutions: " + solutions.length());
      System.out.println("Time: " + (end-start+500)/1000 + " s.");
    }
  }

  /**************************************************************************
   * Print usage information.
   *************************************************************************/
  private static void usage()
  {
    System.out.println("Usage: Wave3 [ -s ] [ <D> ] [ <OPT> ]");
    System.out.println("-s:    silent (no additional information is printed)");
    System.out.println("<D>:   Domain bound (a natural number), default " + D);
    System.out.println("<OPT>: 0 , 1, -1");
    System.out.println("\t 0: solutions satisfying n1+n2=n3 or n1-n2=n3 are printed (default).");
    System.out.println("\t 1: only solutions satisfying n1+n2=n3 are printed.");
    System.out.println("\t-1: only solutions satisfying n1-n2=n3 are printed.");
  }
  
  /**************************************************************************
   * Compute all solutions in domain { (m,n) | m>=0, n>=0, m^2+n^2 <= D^2 } 
   * @param D the domain bound.
   * @param sols in {0,+1,-1} 
   *        = 0: solutions satisfying n1+n2=n3 or n1-n2=n3 are computed.
   *        =-1: only solutions satisfying n1+n2=n3 are computed.
   *        =-2: only solutions satisfying n1-n2=n3 are computed.
   * @returns the solution set.
   *************************************************************************/
  public static SolutionSet solve(int D, int sols)
  {
    int QT = D*D;
    int GT = D;
    int RT = D*D;
    int PT = D;
    PrimeTable primes = new PrimeTable(PT);
    QTable qs = new QTable(QT, primes);
    GammaTable gammas = new GammaTable(GT);
    RootTable roots = new RootTable(RT);
    return new SolutionSet(qs, gammas, roots, D, sols);
  }
  
  // ------------------------------------------------------------------------
  // An integer table of arbitrary size
  // ------------------------------------------------------------------------
  private static class IntTable
  {
    private int[] table = new int[100];
    private int count = 0;
    
    public final void add(int v)
    {
      if (count == table.length) resize();
      table[count] = v;
      count = count+1;
    }
    
    public final int length()
    {
      return count;
    }
    
    public final int get(int i)
    {
      return table[i];
    }
    
    private final void resize()
    {
      int n = table.length;
      int n0 = 2*table.length+1;
      int[] table0 = new int[n0];
      for (int i=0; i<n; i++) table0[i] = table[i];
      table = table0;
    }
    
    public final void print()
    {
      int n = length();
      for (int i=0; i<n; i++)
      {
        System.out.print(get(i));
        if (i+1 < n)
          System.out.print(", ");
        else
          System.out.println();
      }
    }
  }
  
  // ------------------------------------------------------------------------
  // A table for computing integer square roots
  // ------------------------------------------------------------------------
  private final static class RootTable extends IntTable
  {
    /************************************************************************
     * Constructs table for computing the integer square roots
     * of numbers less than or equal max.
     * @param max a bound for the computation.
     **********************************************************************/
    public RootTable(int max)
    {
      int i = 0;
      while (true)
      {
        int s = i*i;
        add(s);
        if (s >= max) break;
        i = i+1;
      }
    }
    
    /************************************************************************
     * Computes the integer square root of x.
     * @param x a number such that 0 <= x <= max.
     * @return the greatest y such that y*y <= x.
     ***********************************************************************/
    public int find(int x)
    {
      int n = length();
      int l = 0;
      int r = n-1;
      while (l <= r)
      {
        int m = (l+r)/2;
        int v = get(m);
        if (v == x) return m;
        if (v < x)
        {
          l = m+1;
          if (l < n && get(l) > x) return m;
        }
        else
        {
          r = m-1;
          if (r >= 0 && get(r) < x) return m-1;
        }
      }
      throw new RuntimeException("invalid square root: " + x);
    }
  }
  
  // -------------------------------------------------------------------------
  // a table of prime numbers
  // -------------------------------------------------------------------------
  private final static class PrimeTable extends IntTable
  {
    /*************************************************************************
     * Constructs a table of all primes less than or equal max.
     * @param max a bound value.
     ************************************************************************/
    public PrimeTable(int max)
    {
      int p = 3;
      add(2); add(3);
      while (p < max) 
      {
        int p0 = next(p);
        add(p0);
        p = p0;
      }
    }
    
    /**************************************************************************
     * Finds the next prime number greater than p.
     * Assumes that all primes less than or equal p have been computed
     * @param p an odd number such that all primes <= p have been computed.
     * @return the next prime number.
     ************************************************************************/
    private int next(int p)
    {
      do { p = p+2; } while (!isPrime(p));
      return p;
    }
    
    /*************************************************************************
     * Tests whether p is prime assuming
     * that all primes less than p have been computed.
     * @param p a number such that all primes less than p have been computed.
     * @return true if and only if p is prime.
     ************************************************************************/
    private boolean isPrime(int p)
    {
      int n = length();
      for (int i=0; i<n; i++)
      {
        if (p % get(i) == 0) return false;
      }
      return true;
    }    
  }
    
  // -------------------------------------------------------------------------
  // a table of q-values
  // -------------------------------------------------------------------------
  private final static class QTable extends IntTable
  {
    /*************************************************************************
     * Construct a table of q-values from 1 to max.
     * @param max an upper bound of the q-values.
     * @param primes a table of prime numbers up to the square root of max.
     ************************************************************************/
    public QTable(int max, PrimeTable primes)
    {
      for (int q=1; q<=max; q++)
      {
        if (squareFreeExtended(q, primes)) add(q);
      }
    }
    
    /*************************************************************************
     * An extended squarefree check.
     * @param n number to be tested
     * @param primes a table of prime numbers up to the square root of n.
     * @return true if and only if n is squarefree and has no prime factor p
     *         such that p % 4 = 3.
     ************************************************************************/
    private boolean squareFreeExtended(int n, PrimeTable primes)
    {
      int i = 0;
      while (true)
      {
        int p = primes.get(i);
        if (p*p > n) return true;
        if (n % p == 0)
        {
          if (p % 4 == 3) return false;
          int q = n/p;
          if (q % 4 == 3) return false;
          if (q % p == 0) return false;
        }
        i = i+1;
      }
    }
  }   
  
  // -------------------------------------------------------------------------
  // a table of gammas
  // -------------------------------------------------------------------------
  private final static class GammaTable
  {
    private IntTable gamma1 = new IntTable();
    private IntTable gamma2 = new IntTable();
    private IntTable gamma3 = new IntTable();
    
    /*************************************************************************
     * Constructs a table of all gamma factors from 1 to max
     * @param max the bound for the factors.
     ************************************************************************/
    public GammaTable(int max)
    {
      for (int g1 = 1; g1 <= max; g1++)
      {
        for (int g2 = 1; g2 <= max; g2++)
        {
          int p = g1*g2;
          int s = g1+g2;
          if (p % s != 0) continue;
          int g3= p/s;
          gamma1.add(g1);
          gamma2.add(g2);
          gamma3.add(g3);
        }
      }
    }
    
    /**************************************************************************
     * Gets the number of table entries.
     * @return the number of table entries.
     *************************************************************************/
    public int length() { return gamma1.length(); }
    
    /**************************************************************************
     * Get the component of the entries
     * @param i an index with 0 <= i < length()
     * @return a component of the entry
     *************************************************************************/
    public int gamma1(int i) { return gamma1.get(i); }
    public int gamma2(int i) { return gamma2.get(i); }
    public int gamma3(int i) { return gamma3.get(i); }
    
    /**************************************************************************
     * Print the table to the output stream.
     *************************************************************************/
    public void print()
    {
      int n = length();
      for (int i=0; i<n; i++)
      {
        System.out.print("[" + 
           gamma1(i) + " " + gamma2(i) + " " + gamma3(i) + "]");
        if (i+1 < n) 
          System.out.print(", ");
        else
          System.out.println();
      }
    }
  }    
  
  // -------------------------------------------------------------------------
  // a pair of integer tables
  // -------------------------------------------------------------------------  
  private final static class IntTablePair
  {
    public final IntTable m; 
    public final IntTable n;
    
    public IntTablePair()
    {
      this.m = new IntTable();
      this.n = new IntTable();
    }
    
    public int length()
    {
      return m.length();
    }
  }
  
  // -------------------------------------------------------------------------
  // a table of solutions
  // -------------------------------------------------------------------------
  public final static class SolutionSet
  {
    // the tables of solution components
    private IntTable m1t = new IntTable();
    private IntTable n1t = new IntTable();
    private IntTable m2t = new IntTable();
    private IntTable n2t = new IntTable();
    private IntTable m3t = new IntTable();
    private IntTable n3t = new IntTable();
      
    /**************************************************************************
     * Constructs set of solution in domain denoted by D.
     * @param qtable   the table of q-values
     * @param gtable   the table of gammas
     * @param roots    the table of roots
     * @param D        the domain bound
     * @param sols     constraint on which solutions are computed
     *************************************************************************/
    public SolutionSet(QTable qtable, GammaTable gtable, RootTable roots, 
      int D, int sols)
    {
     long bound = D*D;
     int qn = qtable.length();
     for (int i=0; i<qn; i++)
     {
       int q = qtable.get(i);
       int gn = gtable.length();
       for (int j=0; j<gn; j++)
       {
         long g1 = gtable.gamma1(j);
         long v1 = q*g1*g1;
         if (v1 > bound) continue;
         long g2 = gtable.gamma2(j);
         long v2 = q*g2*g2;
         if (v2 > bound) continue;
         long g3 = gtable.gamma3(j);
         long v3 = q*g3*g3;
         if (v3 > bound) continue;
         IntTablePair p1 = decompose((int)v1, roots);
         if (p1 == null) continue;
         IntTablePair p2 = decompose((int)v2, roots);
         if (p2 == null) continue;
         IntTablePair p3 = decompose((int)v3, roots);
         if (p3 == null) continue;
         combine(p1, p2, p3, sols);
       }
     }
    }
  
    /*************************************************************************
     * Decompose x into a sum of squares.
     * @param x the number to be composed.
     * @param roots the table of roots used for the computation.
     * @return a table of the solutions.
     ************************************************************************/
    private IntTablePair decompose(int x, RootTable roots)
    {
      IntTablePair result = null;
      int bound = roots.find(x);
      for (int m=0; m<=bound; m++)
      {
        int m2 = m*m;
        int n2 = x-m2;
        int n = roots.find(n2);
        if (n*n != n2) continue;
        if (result == null) result = new IntTablePair();
        result.m.add(m);
        result.n.add(n);
      }
      return result;
    }
    
    /**************************************************************************
     * Add to the solution set all legal combinations of the partial solutions.
     * @param p1 the solutions for (m1,n1)
     * @param p2 the solutions for (m2,n2)
     * @param p3 the solutions for (m3,n3)
     * @param sols constraint on which solutions are computed
     *************************************************************************/
    private void combine(IntTablePair p1, IntTablePair p2, IntTablePair p3,
      int sols)
    {
      Set<String> solutions = new HashSet<String>();
      int l1 = p1.length();
      int l2 = p2.length();
      int l3 = p3.length();
      for (int i=0; i<l1; i++)
      {
        int m1 = p1.m.get(i); 
        int n1 = p1.n.get(i);
        for (int j=0; j<l2; j++)
        {
          int m2 = p2.m.get(j); 
          int n2 = p2.n.get(j);
          for (int k=0; k<l3; k++)
          {
            int m3 = p3.m.get(k); 
            int n3 = p3.n.get(k);
            switch (sols)
            {
              case SOLUTIONSALL:
              {
                if (n1+n2 != n3 && n1-n2 != n3) continue;
                break;
              }
              case SOLUTIONSPLUS:
              {
                if (n1+n2 != n3) continue;
                break;
              }
              case SOLUTIONSMINUS:
              {
                if (n1-n2 != n3) continue;
                break;
              }
              default:
                throw new RuntimeException("invalid solutions option: " + sols);
            }
            // rely on hash table for occurrence check
            String s = m1 + "," + n1 + "," + m2 + "," + n2 + "," + m3 + "," + n3;
            if (solutions.contains(s)) continue;
            solutions.add(s);
            m1t.add(m1); n1t.add(n1);
            m2t.add(m2); n2t.add(n2);
            m3t.add(m3); n3t.add(n3);
          }
        }
      }
    }
    
    /***********************************************************************
     * Get the number of solutions.
     * @return the number of solutions
     **********************************************************************/
    public int length() { return m1t.length(); }
    
    /***********************************************************************
     * Get the solution components.
     * @param i the solution index with 0 <= i < length()
     * @return a component of the solution denoted by i.
     **********************************************************************/
    public int m1(int i) { return m1t.get(i); }
    public int n1(int i) { return n1t.get(i); }
    public int m2(int i) { return m2t.get(i); }
    public int n2(int i) { return n2t.get(i); }
    public int m3(int i) { return m3t.get(i); }
    public int n3(int i) { return n3t.get(i); }
    
    /***********************************************************************
     * Prints the solution set on the standard output stream
     * as a single line.
     **********************************************************************/
    public void print()
    {
      int n = length();
      System.out.print("{");
      for (int i=0; i<n; i++)
      {
        System.out.print("Solution[{" + 
            m1t.get(i) + "," + n1t.get(i) + "}, {" +
            m2t.get(i) + "," + n2t.get(i) + "}, {" +
            m3t.get(i) + "," + n3t.get(i) + "}]");
        if (i+1 < n) System.out.print(", ");
      }
      System.out.println("}");
    }
    
    /***********************************************************************
     * Prints the solution set on the standard output stream
     * as comma-separated values
     **********************************************************************/
    public void printCSV()
    {
      int n = length();
      System.out.println("Index,m1,n1,m2,n2,m3,n3");
      for (int i=0; i<n; i++)
      {
        System.out.println(i + "," +
            m1t.get(i) + "," + n1t.get(i) + "," +
            m2t.get(i) + "," + n2t.get(i) + "," +
            m3t.get(i) + "," + n3t.get(i));
      }
    }
  }
}
