// ----------------------------------------------------------------------------
// Wave4.java
// 4-Wave Resonance Condition Solver
// $Id: Wave4.java,v 1.11 2009/07/03 11:15:30 schreine Exp $
//
// Author: Wolfgang Schreiner <Wolfgang.Schreiner@risc.uni-linz.ac.at>
// Copyright (C) 2009, Research Institute for Symbolic Computation (RISC)
// Johannes Kepler University, Linz, Austria, http://www.risc.uni-linz.ac.at
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.
// ----------------------------------------------------------------------------

import java.util.*;

public final class Wave4
{
  // domain bound to be used for testing
  private final static int D=10;
  
  // print format 
  private final static int FORMATNONE = 0; // no output
  private final static int FORMATLINE = 1; // single line Mathematica format
  private final static int FORMATCSV = 2;  // comma separated values
  private final static int format = FORMATLINE;
  
  // which solutions are printed
  private final static int SOLUTIONSALL   =  0; 
  private final static int SOLUTIONSANGLE = +1; // only angle resonances
  private final static int SOLUTIONSSCALE = -1; // only scale resonances
  
  /***************************************************************************
   * Program to solve resonance conditions for 4-wave interactions.
   * 
   * Call as "java -server Wave4 [ -s ] [ <D> ][ <OPT> ]"
   * 
   * Computes all solutions in domain { (m,n) | m^2+n^2 <= D^2 }
   * 
   * OPT = 0: all solutions are printed (default).
   * OPT = 1: only angle resonances are printed.
   * OPT = 2: only scale resonances are printed.
   * 
   * Prints list of solutions to standard output stream; unless option "-s"
   * is given, also the number of solutions and the time for their 
   * computation is printed.
   * 
   * @param argv the argument vector described above.
   **************************************************************************/
  public static void main(String argv[])
  {
    // input processing
    boolean silent = false;
    int d = D;    
    int n = argv.length;
    int sols = SOLUTIONSALL;
    int p = 0;
    try
    {
      if (p < n && argv[p].equals("-s"))
      {
        silent = true;
        p = p+1;
      }
      if (p < n)
      {
        d = Integer.parseInt(argv[p]);
        if (d < 0) usage();
        p = p+1;
      }
      if (p < n)
      {
        sols = Integer.parseInt(argv[p]);
        if (sols < -1 || sols > +1) usage();
      }
    }
    catch (NumberFormatException e)
    {
      usage();
      return;
    }
    
    // computation
    long start = System.currentTimeMillis();
    SolutionSet solutions = solve(d, sols);
    long end = System.currentTimeMillis();
    
    // output
    if (format == FORMATNONE)
    {       
    }
    else if (format == FORMATLINE)
      solutions.print();
    else if (format == FORMATCSV)
      solutions.printCSV();
    if (!silent)
    {
      System.out.println("Non-collinear Solutions: " + solutions.length());
      System.out.println("Collinear Solutions: " + solutions.getCollinear());
      System.out.println("Time: " + (end-start+500)/1000 + " s.");
    }
  }

  /**************************************************************************
   * Print usage information.
   *************************************************************************/
  private static void usage()
  {
    System.out.println("Usage: Wave4 [ -s ] [ <D> ]");
    System.out.println("-s:   silent (no additional information is printed)");
    System.out.println("<D>:  Domain bound (a natural number), default " + D);
    System.out.println("<OPT>: 0 , 1, -1");
    System.out.println("\t 0: all solutions printed (default).");
    System.out.println("\t 1: only angle-resonances are printed.");
    System.out.println("\t-1: only scale-resonances are printed.");
  }
  
  /**************************************************************************
   * Compute all solutions in domain { (m,n) | m^2+n^2 <= D^2 } 
   * @param D the domain bound.
   * @param sols in {0,+1,-1} 
   *    = 0: all solutions computed.
   *    =-1: only solutions for angle resonances are computed.
   *    =-2: only solutions for scale resonances are computed.
   * @returns the solution set.
   *************************************************************************/
  public static SolutionSet solve(int D, int sols)
  {
    int QT = D*D;
    int GT = (int)Math.sqrt(D);
    int RT = D*D;
    int PT = D;
    PrimeTable primes = new PrimeTable(PT);
    QTable qs = new QTable(QT, primes);
    GammaTable gammas = new GammaTable(GT);
    RootTable roots = new RootTable(RT);
    return new SolutionSet(qs, gammas, roots, D, sols);
  }
  
  // ------------------------------------------------------------------------
  // An integer table of arbitrary size
  // ------------------------------------------------------------------------
  private static class IntTable
  {
    private int[] table = new int[100];
    private int count = 0;
    
    public final void add(int v)
    {
      if (count == table.length) resize();
      table[count] = v;
      count = count+1;
    }
    
    public final int length()
    {
      return count;
    }
    
    public final int get(int i)
    {
      return table[i];
    }
    
    private final void resize()
    {
      int n = table.length;
      int n0 = 2*table.length+1;
      int[] table0 = new int[n0];
      for (int i=0; i<n; i++) table0[i] = table[i];
      table = table0;
    }
    
    public final void print()
    {
      int n = length();
      for (int i=0; i<n; i++)
      {
        System.out.print(get(i));
        if (i+1 < n)
          System.out.print(", ");
        else
          System.out.println();
      }
    }
  }
     
  // ------------------------------------------------------------------------
  // A table for computing integer square roots
  // ------------------------------------------------------------------------
  private final static class RootTable extends IntTable
  {
    /************************************************************************
     * Constructs table for computing the integer square roots
     * of numbers less than or equal max.
     * @param max a bound for the computation.
     **********************************************************************/
    public RootTable(int max)
    {
      int i = 0;
      while (true)
      {
        int s = i*i;
        add(s);
        if (s >= max) break;
        i = i+1;
      }
    }
    
    /************************************************************************
     * Computes the integer square root of x.
     * @param x a number such that 0 <= x <= max.
     * @return the greatest y such that y*y <= x.
     ***********************************************************************/
    public int find(int x)
    {
      int n = length();
      int l = 0;
      int r = n-1;
      while (l <= r)
      {
        int m = (l+r)/2;
        int v = get(m);
        if (v == x) return m;
        if (v < x)
        {
          l = m+1;
          if (l < n && get(l) > x) return m;
        }
        else
        {
          r = m-1;
          if (r >= 0 && get(r) < x) return m-1;
        }
      }
      throw new RuntimeException("invalid square root: " + x);
    }
  }
  
  // -------------------------------------------------------------------------
  // a table of prime numbers
  // -------------------------------------------------------------------------
  private final static class PrimeTable extends IntTable
  {
    /*************************************************************************
     * Constructs a table of all primes less than or equal max.
     * @param max a bound value.
     ************************************************************************/
    public PrimeTable(int max)
    {
      int p = 3;
      add(2); add(3);
      while (p < max) 
      {
        int p0 = next(p);
        add(p0);
        p = p0;
      }
    }
    
    /**************************************************************************
     * Finds the next prime number greater than p.
     * Assumes that all primes less than or equal p have been computed
     * @param p an odd number such that all primes <= p have been computed.
     * @return the next prime number.
     ************************************************************************/
    private int next(int p)
    {
      do { p = p+2; } while (!isPrime(p));
      return p;
    }
    
    /*************************************************************************
     * Tests whether p is prime assuming
     * that all primes less than p have been computed.
     * @param p a number such that all primes less than p have been computed.
     * @return true if and only if p is prime.
     ************************************************************************/
    private boolean isPrime(int p)
    {
      int n = length();
      for (int i=0; i<n; i++)
      {
        if (p % get(i) == 0) return false;
      }
      return true;
    }    
  }
    
  // -------------------------------------------------------------------------
  // a table of q-values
  // -------------------------------------------------------------------------
  private final static class QTable extends IntTable
  {
    /*************************************************************************
     * Construct a table of q-values from 1 to max.
     * @param max an upper bound of the q-values.
     * @param primes a table of prime numbers up to the square root of max.
     ************************************************************************/
    public QTable(int max, PrimeTable primes)
    {
      for (int q=1; q<=max; q++)
      {
        if (quadFreeExtended(q, primes)) add(q);
      }
    }
    
    
    /*************************************************************************
     * An extended quad-free check.
     * @param n number to be tested
     * @param primes a table of prime numbers up to the square root of n.
     * @return true if and only if 
     *         - n has no factors of degree 4 and
     *         - all factors of n with form 4u+3 have degree 2
     ************************************************************************/
    private boolean quadFreeExtended(int n, PrimeTable primes)
    {
      int i = 0;
      while (true)
      {
        int p = primes.get(i);
        if (p*p > n) return true;
        if (n % p == 0)
        {
          int n1 = n/p;
          if (p % 4 == 3) 
          {
            if (n1 % p != 0 || (n1/p) % p == 0) return false;
          }
          else
          {
            if (n1 % p == 0)
            {
              int n2 = n1/p;
              if (n2 % p == 0)
              {
                int n3 = n2/p;
                if (n3 % p == 0) return false;
              }
            }
          }
        }
        i = i+1;
      }
    }
  }   
  
  // -------------------------------------------------------------------------
  // a table of gammas
  // -------------------------------------------------------------------------
  private final static class GammaTable
  {
    private IntTable gamma1 = new IntTable();
    private IntTable gamma2 = new IntTable();
    private IntTable gamma3 = new IntTable();
    private IntTable gamma4 = new IntTable();
    
    /*************************************************************************
     * Constructs a table of all gamma factors from 1 to max
     * @param max the bound for the factors.
     ************************************************************************/
    public GammaTable(int max)
    {
      // make use of symmetry to only consider certain combinations
      for (int g1 = 1; g1 <= max; g1++)
      {
        // g1 <= g2
        for (int g2 = g1; g2 <= max; g2++)
        {
          int gs = g1+g2;
          // g3 <= g4
          for (int g3 = 1; g3 <= gs/2; g3++)
          {
            int g4 = gs-g3;
            // g1,g2 <-> g3/g4
            if (contains(g3,g4,g1,g2)) continue;
            gamma1.add(g1);
            gamma2.add(g2);
            gamma3.add(g3);
            gamma4.add(g4);
          }
        }
      }
    }
    
    /************************************************************************
     * Returns true if table contains denoted gamma-values.
     * @param g1-g4 the gamma values.
     * @return true if and only if g1-g4 are contained in the table.
     ***********************************************************************/
    private boolean contains(int g1, int g2, int g3, int g4)
    {
      int n = length();
      for (int i=0; i<n; i++)
      {
        if (gamma1(i) != g1) continue;
        if (gamma2(i) != g2) continue;
        if (gamma3(i) != g3) continue;
        if (gamma4(i) != g4) continue;
        return true;
      }
      return false;
    }
    
    /**************************************************************************
     * Gets the number of table entries.
     * @return the number of table entries.
     *************************************************************************/
    public int length() { return gamma1.length(); }
    
    /**************************************************************************
     * Get the component of the entries
     * @param i an index with 0 <= i < length()
     * @return a component of the entry
     *************************************************************************/
    public int gamma1(int i) { return gamma1.get(i); }
    public int gamma2(int i) { return gamma2.get(i); }
    public int gamma3(int i) { return gamma3.get(i); }
    public int gamma4(int i) { return gamma4.get(i); }
    
    /**************************************************************************
     * Print the table to the output stream.
     *************************************************************************/
    public void print()
    {
      int n = length();
      for (int i=0; i<n; i++)
      {
        System.out.print("[" + 
           gamma1(i) + " " + gamma2(i) + " " + gamma3(i) + " " + gamma4(i) + "]");
        if (i+1 < n) 
          System.out.print(", ");
        else
          System.out.println();
      }
    }
  }    
  
  // -------------------------------------------------------------------------
  // a pair of integer tables
  // -------------------------------------------------------------------------  
  private final static class IntTablePair
  {
    public final IntTable m; 
    public final IntTable n;
    
    public IntTablePair()
    {
      this.m = new IntTable();
      this.n = new IntTable();
    }
    
    public int length()
    {
      return m.length();
    }
  }
  
  // -------------------------------------------------------------------------
  // a table of solutions
  // -------------------------------------------------------------------------
  public final static class SolutionSet
  {
    // the tables of solution components
    // representing all non-collinear solutions
    private IntTable m1t = new IntTable();
    private IntTable n1t = new IntTable();
    private IntTable m2t = new IntTable();
    private IntTable n2t = new IntTable();
    private IntTable m3t = new IntTable();
    private IntTable n3t = new IntTable();
    private IntTable m4t = new IntTable();
    private IntTable n4t = new IntTable();
    
    // the number of collinear solutions
    private int collinear = 0;
    
    /**************************************************************************
     * Constructs set of solution in domain denoted by D.
     * @param qtable   the table of q-values
     * @param gtable   the table of gammas
     * @param roots    the table of roots
     * @param D        the domain bound
     * @param sols     constraint on which solutions are computed
     *************************************************************************/
    public SolutionSet(QTable qtable, GammaTable gtable, RootTable roots, 
      int D, int sols)
    {
     long bound = D*D;
     int qn = qtable.length();
     for (int i=0; i<qn; i++)
     {
       int q = qtable.get(i);
       int gn = gtable.length();
       for (int j=0; j<gn; j++)
       {
         long g1 = gtable.gamma1(j);
         long h1 = g1*g1;
         long v1 = q*h1*h1;
         if (v1 > bound) continue;
         long g2 = gtable.gamma2(j);
         long h2 = g2*g2;
         long v2 = q*h2*h2;
         if (v2 > bound) continue;
         long g3 = gtable.gamma3(j);
         long h3 = g3*g3;
         long v3 = q*h3*h3;
         if (v3 > bound) continue;
         long g4 = gtable.gamma4(j);
         long h4 = g4*g4;
         long v4 = q*h4*h4;
         if (v4 > bound) continue;
         IntTablePair p1 = decompose((int)v1, roots);
         if (p1 == null) continue;
         IntTablePair p2 = decompose((int)v2, roots);
         if (p2 == null) continue;
         IntTablePair p3 = decompose((int)v3, roots);
         if (p3 == null) continue;
         IntTablePair p4 = decompose((int)v4, roots);
         if (p4 == null) continue;
         combine(p1, p2, p3, p4, sols);
       }
     }
    }
  
    /*************************************************************************
     * Decompose x into a sum of squares.
     * @param x the number to be composed.
     * @param roots the table of roots used for the computation.
     * @param D the bound for the domain in which the solution shall be
     * @return a table of the solutions.
     ************************************************************************/
    private IntTablePair decompose(int x, RootTable roots)
    {
      IntTablePair result = null;
      int bound = roots.find(x);
      for (int m=0; m<=bound; m++)
      {
        int m2 = m*m;
        int n2 = x-m2;
        int n = roots.find(n2);
        if (n*n != n2) continue;
        if (result == null) result = new IntTablePair();
        add(result, m, n);
      }
      return result;
    }
    
    /*************************************************************************
     * Add solutions derived from (m,n) to solution set.
     * @param result the solution set.
     * @param m non-negative first component of a solution.
     * @param n non-negative second component of a solution.
     ************************************************************************/
    private void add(IntTablePair result, int m, int n)
    {
      result.m.add(m); result.n.add(n);
      if (m == 0)
      {
        if (n != 0)
        {
          result.m.add(m); result.n.add(-n);
        }
      }
      else
      {
        result.m.add(-m); result.n.add(n);
        if (n != 0)
        {
          result.m.add(m); result.n.add(-n);
          result.m.add(-m); result.n.add(-n);
        }
      }
    }
    
    /**************************************************************************
     * Add to the solution set all legal combinations of the partial solutions.
     * @param p1 the solutions for (m1,n1)
     * @param p2 the solutions for (m2,n2)
     * @param p3 the solutions for (m3,n3)
     * @param p4 the solutions for (m4,n4)
     * @param sols constraint on which solutions are computed
     *************************************************************************/
    private void combine(IntTablePair p1, IntTablePair p2, 
      IntTablePair p3, IntTablePair p4, int sols)
    {
      Set<String> solutions = new HashSet<String>();
      CanonicalSolution solution = new CanonicalSolution();
      int l1 = p1.length();
      int l2 = p2.length();
      int l3 = p3.length();
      int l4 = p4.length();
      for (int i=0; i<l1; i++)
      {
        int m1 = p1.m.get(i); 
        int n1 = p1.n.get(i);
        for (int j=0; j<l2; j++)
        {
          int m2 = p2.m.get(j); 
          int n2 = p2.n.get(j);
          for (int k=0; k<l3; k++)
          {
            int m3 = p3.m.get(k); 
            int n3 = p3.n.get(k);
            for (int l=0; l<l4; l++)
            {
              int m4 = p4.m.get(l); 
              int n4 = p4.n.get(l);
              if (m1+m2 != m3+m4 || n1+n2 != n3+n4) continue;
              if (collinear(m1, n1, m2, n2, m3, n3, m4, n4))
              {
                collinear = collinear+1;
                continue;
              }
              switch (sols)
              {
                case SOLUTIONSALL:
                {
                  break;
                }
                case SOLUTIONSANGLE:
                {
                  if (!angle(m1, n1, m2, n2, m3, n3, m4, n4)) continue;
                  break;
                }
                case SOLUTIONSSCALE:
                {
                  if (angle(m1, n1, m2, n2, m3, n3, m4, n4)) continue;
                  break;
                }
                default:
                  throw new RuntimeException("invalid solutions option: " + sols);
              }
              boolean okay = solution.canonize(m1, n1, m2, n2, m3, n3, m4, n4);
              if (!okay) continue;
              int m10 = solution.m1(); int n10 = solution.n1();
              int m20 = solution.m2(); int n20 = solution.n2();
              int m30 = solution.m3(); int n30 = solution.n3();
              int m40 = solution.m4(); int n40 = solution.n4();
              String s1 = "" + m10 + "," + n10 + "," + m20 + "," + n20 + ","
                             + m30 + "," + n30 + "," + m40 + "," + n40;
              // rely on hash table for occurrence check
              if (solutions.contains(s1)) continue;
              solutions.add(s1);
              // variables may be switched
              String s2 = "" + n10 + "," + m10 + "," + n20 + "," + m20 + ","
                             + n30 + "," + m30 + "," + n40 + "," + m40;
              solutions.add(s2);
              m1t.add(m10); n1t.add(n10);
              m2t.add(m20); n2t.add(n20);
              m3t.add(m30); n3t.add(n30);
              m4t.add(m40); n4t.add(n40);
            }
          }
        }
      }
    }
    
    /**************************************************************************
     * Check whether four vectors are collinear.
     * @param m1..n4 the components of the vectors.
     * @return true iff all vectors are collinear.
     *************************************************************************/
    private boolean collinear(int m1, int n1, int m2, int n2,
      int m3, int n3, int m4, int n4)
    {
      if (Math.abs(m1*n2) != Math.abs(n1*m2)) return false;
      if (Math.abs(m1*n3) != Math.abs(n1*m3)) return false;
      if (Math.abs(m1*n4) != Math.abs(n1*m4)) return false;
      return true;
    }
    
    /***********************************************************************
     * Check whether solution denotes an angle resonance, i.e. there are at
     * most 2 different vector lengths.
     * @return true iff solution denotes an angle resonance.
     **********************************************************************/
    private boolean angle(int m1, int n1, int m2, int n2,
      int m3, int n3, int m4, int n4)
    {
      int l1 = m1*m1+n1*n1;
      int l2 = m2*m2+n2*n2;
      int l3 = m3*m3+n3*n3;
      if (l1 == l2)
      {
        if (l3 == l1) return true;
        int l4 = m4*m4+n4*n4;
        return (l4 == l1 || l4 == l3);
      }
      else
      {
        if (l3 != l1 && l3 != l2) return false;
        int l4 = m4*m4+n4*n4;
        return (l4 == l1 || l4 == l2);
      }
    }
    
    /***********************************************************************
     * Get the number of non-collinear solutions.
     * @return the number of non-collinear solutions
     **********************************************************************/
    public int length() { return m1t.length(); }
    
    /***********************************************************************
     * Get the components of the non-collinear solutions.
     * @param i the solution index with 0 <= i < length()
     * @return a component of the non-collinear solution denoted by i.
     **********************************************************************/
    public int m1(int i) { return m1t.get(i); }
    public int n1(int i) { return n1t.get(i); }
    public int m2(int i) { return m2t.get(i); }
    public int n2(int i) { return n2t.get(i); }
    public int m3(int i) { return m3t.get(i); }
    public int n3(int i) { return n3t.get(i); }
    
    /***********************************************************************
     * Get the number of collinear solutions.
     * @return the number of collinear solutions
     **********************************************************************/
    public int getCollinear()
    {
      return collinear;
    }
    
    /***********************************************************************
     * Prints the solution set on the standard output stream
     * as a single line.
     **********************************************************************/
    public void print()
    {
      int n = length();
      System.out.print("{");
      for (int i=0; i<n; i++)
      {
        System.out.print("Solution[{" + 
            m1t.get(i) + "," + n1t.get(i) + "}, {" +
            m2t.get(i) + "," + n2t.get(i) + "}, {" +
            m3t.get(i) + "," + n3t.get(i) + "}, {" +
            m4t.get(i) + "," + n4t.get(i) + "}]");
        if (i+1 < n) System.out.print(", ");
      }
      System.out.println("}");
    }
    
    /***********************************************************************
     * Prints the solution set on the standard output stream
     * as comma-separated values
     **********************************************************************/
    public void printCSV()
    {
      int n = length();
      System.out.println("Index,m1,n1,m2,n2,m3,n3,m4,n4");
      for (int i=0; i<n; i++)
      {
        System.out.println(i + "," +
            m1t.get(i) + "," + n1t.get(i) + "," +
            m2t.get(i) + "," + n2t.get(i) + "," +
            m3t.get(i) + "," + n3t.get(i) + "," +
            m4t.get(i) + "," + n4t.get(i));
      }
    }
  }
  
  // -------------------------------------------------------------------------
  // A canonical form of a solution (m1,n1,m2,n2,m3,n3,m4,n4)
  // * (m1,n1) <= (m2,n2), (m3,n3) <= (m4,n4) 
  // * (m1,n1,m2,n2) < (m3,n3,m4,n4) 
  // -------------------------------------------------------------------------
  private final static class CanonicalSolution
  {
    // a wave vector 
    private final static class WaveVector
    {
      int m; int n;
      boolean equals(WaveVector w)
      {
        return m == w.m && n == w.n;
      }
      boolean less(WaveVector w)
      {
        return m < w.m || (m == w.m && n < w.n);
      }    
      boolean lessEqual(WaveVector w)
      {
        return m < w.m || (m == w.m && n <= w.n);
      }
    }
    
    // a pair of wave vectors 
    private final static class WaveVectorPair
    {
      WaveVector w1 = new WaveVector();
      WaveVector w2 = new WaveVector();
      boolean equals(WaveVectorPair s)
      {
        return w1.equals(s.w1) && w2.equals(s.w2);
      }
      boolean lessEqual(WaveVectorPair s)
      {
        return w1.lessEqual(s.w1) || (w1.equals(s.w1) && w2.lessEqual(w2));
      }
      void sort()
      {
        WaveVector w1 = this.w1;
        WaveVector w2 = this.w2;
        if (w1.lessEqual(w2)) return;
        this.w1 = w2;
        this.w2 = w1;
      }
    }
    
    // the left and the right side of the solution equation
    WaveVectorPair left = new WaveVectorPair();
    WaveVectorPair right = new WaveVectorPair();
    
    /*************************************************************************
     * Construct canonical solution
     * @param m1..n4 the solution components
     * @return true iff solution is valid.
     ************************************************************************/
    public boolean canonize(int m1, int n1, int m2, int n2,
                            int m3, int n3, int m4, int n4)
    {
      left.w1.m  = m1; left.w1.n  = n1;
      left.w2.m  = m2; left.w2.n  = n2;
      right.w1.m = m3; right.w1.n = n3;
      right.w2.m = m4; right.w2.n = n4;
      left.sort(); right.sort();
      if (left.equals(right)) return false;
      sort();
      return true;
    }
    
    /*************************************************************************
     * Get components of canonical solution.
     * @return a component.
     ************************************************************************/
    public int m1() { return left.w1.m; }
    public int n1() { return left.w1.n; }
    public int m2() { return left.w2.m; }
    public int n2() { return left.w2.n; }
    public int m3() { return right.w1.m; }
    public int n3() { return right.w1.n; }
    public int m4() { return right.w2.m; }
    public int n4() { return right.w2.n; }
    
    /**************************************************************************
     * Sort both sides of the solution.
     *************************************************************************/
    private void sort()
    {
      WaveVectorPair left = this.left;
      WaveVectorPair right = this.right;
      if (left.lessEqual(right)) return;
      this.left = right;
      this.right = left;
    }
  }
}

